from typing import Mapping, Any, Optional
from feldera.rest.sql_table import SQLTable
from feldera.rest.sql_view import SQLView


class Pipeline:
    """
    Represents a Feldera pipeline
    """

    def __init__(
        self,
        name: str,
        sql: str,
        udf_rust: str,
        udf_toml: str,
        program_config: Mapping[str, Any],
        runtime_config: Mapping[str, Any],
        description: Optional[str] = None,
    ):
        """
        Initializes a new pipeline

        :param name: The name of the pipeline
        :param sql: The SQL code of the pipeline
        :param udf_rust: Rust code for UDFs
        :param udf_toml: Rust dependencies required by UDFs (in the TOML format)
        :param program_config: The program config of the pipeline
        :param runtime_config: The configuration of the pipeline
        :param description: Optional. The description of the pipeline
        """

        self.name: str = name
        self.program_code: str = sql.strip()
        self.udf_rust: str = udf_rust
        self.udf_toml: str = udf_toml
        self.description: Optional[str] = description
        self.program_config: Mapping[str, Any] = program_config
        self.runtime_config: Mapping[str, Any] = runtime_config
        self.id: Optional[str] = None
        self.tables: list[SQLTable] = []
        self.views: list[SQLView] = []
        self.deployment_status: Optional[str] = None
        self.deployment_status_since: Optional[str] = None
        self.created_at: Optional[str] = None
        self.version: Optional[int] = None
        self.program_version: Optional[int] = None
        self.deployment_config: Optional[dict] = None
        self.deployment_desired_status: Optional[str] = None
        self.deployment_error: Optional[dict] = None
        self.deployment_location: Optional[str] = None
        self.program_binary_url: Optional[str] = None
        self.program_info: Optional[dict] = (
            None  # info about input & output connectors and the schema
        )
        self.program_status: Optional[str] = None
        self.program_status_since: Optional[str] = None
        self.program_error: Optional[dict] = None

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        pipeline = cls("", "", "", "", {}, {})
        pipeline.__dict__ = d
        pipeline.tables = []
        pipeline.views = []

        info = d.get("program_info")

        if info is not None:
            for i in info["schema"]["inputs"]:
                tbl = SQLTable.from_dict(i)
                pipeline.tables.append(tbl)

            for output in info["schema"]["outputs"]:
                v = SQLView.from_dict(output)
                pipeline.views.append(v)

        return pipeline



import logging

from feldera.rest.config import Config

from feldera.rest.errors import (
    FelderaAPIError,
    FelderaTimeoutError,
    FelderaCommunicationError,
)

import json
import requests
from requests.packages import urllib3
from typing import Callable, Optional, Any, Union, Mapping, Sequence, List


def json_serialize(body: Any) -> str:
    # serialize as string if this object cannot be serialized (e.g. UUID)
    return json.dumps(body, default=str) if body else "" if body == "" else "null"


class HttpRequests:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.headers = {"User-Agent": "feldera-python-sdk/v1"}
        self.requests_verify = config.requests_verify

        if not self.requests_verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        if self.config.api_key:
            self.headers["Authorization"] = f"Bearer {self.config.api_key}"

    def send_request(
        self,
        http_method: Callable,
        path: str,
        body: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]], List[str], str]
        ] = None,
        content_type: str = "application/json",
        params: Optional[Mapping[str, Any]] = None,
        stream: bool = False,
        serialize: bool = True,
    ) -> Any:
        """
        :param http_method: The HTTP method to use. Takes the equivalent `requests.*` module. (Example: `requests.get`)
        :param path: The path to send the request to.
        :param body: The HTTP request body.
        :param content_type: The value for `Content-Type` HTTP header. "application/json" by default.
        :param params: The query parameters part of this request.
        :param stream: True if the response is expected to be a HTTP stream.
        :param serialize: True if the body needs to be serialized to JSON.
        """
        self.headers["Content-Type"] = content_type

        try:
            timeout = self.config.timeout
            headers = self.headers

            request_path = self.config.url + "/" + self.config.version + path

            logging.debug(
                "sending %s request to: %s with headers: %s, and params: %s",
                http_method.__name__,
                request_path,
                str(headers),
                str(params),
            )

            if http_method.__name__ == "get":
                request = http_method(
                    request_path,
                    timeout=timeout,
                    headers=headers,
                    params=params,
                    stream=stream,
                    verify=self.requests_verify,
                )
            elif isinstance(body, bytes):
                request = http_method(
                    request_path,
                    timeout=timeout,
                    headers=headers,
                    data=body,
                    params=params,
                    stream=stream,
                    verify=self.requests_verify,
                )
            else:
                request = http_method(
                    request_path,
                    timeout=timeout,
                    headers=headers,
                    data=json_serialize(body) if serialize else body,
                    params=params,
                    stream=stream,
                    verify=self.requests_verify,
                )

            resp = self.__validate(request, stream=stream)
            logging.debug("got response: %s", str(resp))
            return resp

        except requests.exceptions.Timeout as err:
            raise FelderaTimeoutError(str(err)) from err
        except requests.exceptions.ConnectionError as err:
            raise FelderaCommunicationError(str(err)) from err

    def get(
        self,
        path: str,
        params: Optional[Mapping[str, Any]] = None,
        stream: bool = False,
    ) -> Any:
        return self.send_request(requests.get, path, params=params, stream=stream)

    def post(
        self,
        path: str,
        body: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]], List[str], str]
        ] = None,
        content_type: Optional[str] = "application/json",
        params: Optional[Mapping[str, Any]] = None,
        stream: bool = False,
        serialize: bool = True,
    ) -> Any:
        return self.send_request(
            requests.post,
            path,
            body,
            content_type,
            params,
            stream=stream,
            serialize=serialize,
        )

    def patch(
        self,
        path: str,
        body: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]], List[str], str]
        ] = None,
        content_type: Optional[str] = "application/json",
        params: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        return self.send_request(requests.patch, path, body, content_type, params)

    def put(
        self,
        path: str,
        body: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]], List[str], str]
        ] = None,
        content_type: Optional[str] = "application/json",
        params: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        return self.send_request(requests.put, path, body, content_type, params)

    def delete(
        self,
        path: str,
        body: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]], List[str]]
        ] = None,
        params: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        return self.send_request(requests.delete, path, body, params=params)

    @staticmethod
    def __to_json(request: requests.Response) -> Any:
        if request.content == b"":
            return request
        return request.json()

    @staticmethod
    def __validate(request: requests.Response, stream=False) -> Any:
        try:
            request.raise_for_status()

            if stream:
                return request
            if request.headers.get("content-type") == "text/plain":
                return request.text
            elif request.headers.get("content-type") == "application/octet-stream":
                return request.content

            resp = HttpRequests.__to_json(request)
            return resp
        except requests.exceptions.HTTPError as err:
            raise FelderaAPIError(str(err), request) from err



from enum import Enum
from typing import Optional


class CompilationProfile(Enum):
    """
    The compilation profile to use when compiling the program.
    """

    SERVER_DEFAULT = None
    """
    The compiler server default compilation profile.
    """

    DEV = "dev"
    """
    The development compilation profile.
    """

    UNOPTIMIZED = "unoptimized"
    """
    The unoptimized compilation profile.
    """

    OPTIMIZED = "optimized"
    """
    The optimized compilation profile, the default for this API.
    """


class BuildMode(Enum):
    CREATE = 1
    GET = 2
    GET_OR_CREATE = 3


class PipelineStatus(Enum):
    """
    Represents the state that this pipeline is currently in.

    .. code-block:: text

        Shutdown     ◄────┐
        │         │
        /deploy   │       │
        │   ⌛ShuttingDown
        ▼         ▲
        ⌛Provisioning    │
        │         │
        Provisioned        │
        ▼         │/shutdown
        ⌛Initializing     │
        │        │
        ┌────────┴─────────┴─┐
        │        ▼           │
        │      Paused        │
        │      │    ▲        │
        │/start│    │/pause  │
        │      ▼    │        │
        │     Running        │
        └──────────┬─────────┘
                   │
                   ▼
                Failed
    """

    NOT_FOUND = 1
    """
    The pipeline has not been created yet.
    """

    SHUTDOWN = 2
    """
    Pipeline has not been started or has been shut down.

    The pipeline remains in this state until the user triggers
    a deployment by invoking the `/deploy` endpoint.
    """

    PROVISIONING = 3
    """
    The runner triggered a deployment of the pipeline and is
    waiting for the pipeline HTTP server to come up.

    In this state, the runner provisions a runtime for the pipeline,
    starts the pipeline within this runtime and waits for it to start accepting HTTP requests.

    The user is unable to communicate with the pipeline during this
    time.  The pipeline remains in this state until:

        1. Its HTTP server is up and running; the pipeline transitions to the
           `PipelineStatus.INITIALIZING` state.
        2. A pre-defined timeout has passed.  The runner performs forced
           shutdown of the pipeline; returns to the `PipelineStatus.SHUTDOWN` state.
        3. The user cancels the pipeline by invoking the `/shutdown` endpoint.
           The manager performs forced shutdown of the pipeline, returns to the
           `PipelineStatus.SHUTDOWN` state.

    """

    INITIALIZING = 4
    """
    The pipeline is initializing its internal state and connectors.

    This state is part of the pipeline's deployment process.  In this state,
    the pipeline's HTTP server is up and running, but its query engine
    and input and output connectors are still initializing.

    The pipeline remains in this state until:

        1.  Initialization completes successfully; the pipeline transitions to the
            `PipelineStatus.PAUSED` state.
        2.  Initialization fails; transitions to the `PipelineStatus.FAILED` state.
        3.  A pre-defined timeout has passed.  The runner performs forced
            shutdown of the pipeline; returns to the `PipelineStatus.SHUTDOWN` state.
        4.  The user cancels the pipeline by invoking the `/shutdown` endpoint.
            The manager performs forced shutdown of the pipeline; returns to the
            `PipelineStatus.SHUTDOWN` state.

    """

    PAUSED = 5
    """
    The pipeline is fully initialized, but data processing has been paused.

    The pipeline remains in this state until:

        1.  The user starts the pipeline by invoking the `/start` endpoint. The
            manager passes the request to the pipeline; transitions to the
            `PipelineStatus.RUNNING` state.
        2.  The user cancels the pipeline by invoking the `/shutdown` endpoint.
            The manager passes the shutdown request to the pipeline to perform a
            graceful shutdown; transitions to the `PipelineStatus.SHUTTING_DOWN` state.
        3.  An unexpected runtime error renders the pipeline `PipelineStatus.FAILED`.

    """

    RUNNING = 6
    """
    The pipeline is processing data.

    The pipeline remains in this state until:

        1. The user pauses the pipeline by invoking the `/pause` endpoint. The
           manager passes the request to the pipeline; transitions to the
           `PipelineStatus.PAUSED` state.
        2. The user cancels the pipeline by invoking the `/shutdown` endpoint.
           The runner passes the shutdown request to the pipeline to perform a
           graceful shutdown; transitions to the
           `PipelineStatus.SHUTTING_DOWN` state.
        3. An unexpected runtime error renders the pipeline
           `PipelineStatus.FAILED`.

    """

    SHUTTING_DOWN = 7
    """
    Graceful shutdown in progress.

    In this state, the pipeline finishes any ongoing data processing,
    produces final outputs, shuts down input/output connectors and
    terminates.

    The pipeline remains in this state until:

        1. Shutdown completes successfully; transitions to the `PipelineStatus.SHUTDOWN` state.
        2. A pre-defined timeout has passed. The manager performs forced shutdown of the pipeline; returns to the
           `PipelineStatus.SHUTDOWN` state.

    """

    FAILED = 8
    """
    The pipeline remains in this state until the users acknowledge the failure
    by issuing a call to shutdown the pipeline; transitions to the
    `PipelineStatus.SHUTDOWN` state.
    """

    UNAVAILABLE = 9
    """
    The pipeline was at least once initialized, but in the most recent status check either
    could not be reached or returned it is not yet ready.
    """

    @staticmethod
    def from_str(value):
        for member in PipelineStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(f"Unknown value '{value}' for enum {PipelineStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value


class ProgramStatus(Enum):
    Pending = 1
    CompilingSql = 2
    SqlCompiled = 3
    CompilingRust = 4
    Success = 5
    SqlError = 6
    RustError = 7
    SystemError = 8

    def __init__(self, value):
        self.error: Optional[dict] = None
        self._value_ = value

    @staticmethod
    def from_value(value):
        error = None
        if isinstance(value, dict):
            error = value
            value = list(value.keys())[0]

        for member in ProgramStatus:
            if member.name.lower() == value.lower():
                member.error = error
                return member
        raise ValueError(f"Unknown value '{value}' for enum {ProgramStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value

    def __str__(self):
        return self.name + (f": ({self.error})" if self.error else "")

    def get_error(self) -> Optional[dict]:
        """
        Returns the compilation error, if any.
        """

        return self.error



class SQLTable:
    """
    Represents a SQL table in Feldera
    """

    def __init__(
        self,
        name: str,
        fields: list[dict],
        case_sensitive: bool = False,
        materialized: bool = False,
    ):
        self.name = name
        self.case_sensitive = case_sensitive
        self.materialized = materialized
        self.fields: list[dict] = fields

    @classmethod
    def from_dict(self, table_dict: dict):
        tbl = SQLTable(name=table_dict["name"], fields=table_dict["fields"])
        tbl.case_sensitive = table_dict["case_sensitive"]
        tbl.materialized = table_dict["materialized"]
        return tbl



from feldera.rest.feldera_client import FelderaClient as FelderaClient
from feldera.pipeline import Pipeline as Pipeline
from feldera.pipeline_builder import PipelineBuilder as PipelineBuilder

import pretty_errors

pretty_errors.configure(
    line_number_first=True,
)

pretty_errors.activate()



from typing import Optional, Any, Mapping


class Resources:
    """
    Class used to specify the resource configuration for a pipeline.

    :param config: A dictionary containing all the configuration values.
    :param cpu_cores_max: The maximum number of CPU cores to reserve for an instance of the pipeline.
    :param cpu_cores_min: The minimum number of CPU cores to reserve for an instance of the pipeline.
    :param memory_mb_max: The maximum memory in Megabytes to reserve for an instance of the pipeline.
    :param memory_mb_min: The minimum memory in Megabytes to reserve for an instance of the pipeline.
    :param storage_class: The storage class to use for the pipeline. The class determines storage performance such
        as IOPS and throughput.
    :param storage_mb_max: The  storage in Megabytes to reserve for an instance of the pipeline.
    """

    def __init__(
        self,
        config: Optional[Mapping[str, Any]] = None,
        cpu_cores_max: Optional[int] = None,
        cpu_cores_min: Optional[int] = None,
        memory_mb_max: Optional[int] = None,
        memory_mb_min: Optional[int] = None,
        storage_class: Optional[str] = None,
        storage_mb_max: Optional[int] = None,
    ):
        config = config or {}

        self.cpu_cores_max = cpu_cores_max
        self.cpu_cores_min = cpu_cores_min
        self.memory_mb_max = memory_mb_max
        self.memory_mb_min = memory_mb_min
        self.storage_class = storage_class
        self.storage_mb_max = storage_mb_max

        self.__dict__.update(config)


class Storage:
    """Storage configuration for a pipeline.

    :param min_storage_bytes: The minimum estimated number of bytes in a batch of data to write it to storage.
    """

    def __init__(
        self,
        config: Optional[Mapping[str, Any]] = None,
        min_storage_bytes: Optional[int] = None,
    ):
        config = config or {}

        self.min_storage_bytes = min_storage_bytes

        self.__dict__.update(config)


class RuntimeConfig:
    """
    Runtime configuration class to define the configuration for a pipeline.
    """

    def __init__(
        self,
        workers: Optional[int] = None,
        storage: Optional[Storage] = None,
        tracing: Optional[bool] = False,
        tracing_endpoint_jaeger: Optional[str] = "",
        cpu_profiler: bool = True,
        max_buffering_delay_usecs: int = 0,
        min_batch_size_records: int = 0,
        clock_resolution_usecs: Optional[int] = None,
        provisioning_timeout_secs: Optional[int] = None,
        resources: Optional[Resources] = None,
    ):
        self.workers = workers
        self.storage = storage
        self.tracing = tracing
        self.tracing_endpoint_jaeger = tracing_endpoint_jaeger
        self.cpu_profiler = cpu_profiler
        self.max_buffering_delay_usecs = max_buffering_delay_usecs
        self.min_batch_size_records = min_batch_size_records
        self.clock_resolution_usecs = clock_resolution_usecs
        self.provisioning_timeout_secs = provisioning_timeout_secs
        if resources is not None:
            self.resources = resources.__dict__

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        """
        Create a `.RuntimeConfig` object from a dictionary.
        """

        conf = cls()
        conf.__dict__ = d
        return conf



from typing import Optional


class Config:
    """
    :class:`.FelderaClient`'s credentials and configuration parameters
    """

    def __init__(
        self,
        url: str,
        api_key: Optional[str] = None,
        version: Optional[str] = None,
        timeout: Optional[float] = None,
        requests_verify: bool = True,
    ) -> None:
        """
        :param url: The url to the Feldera API (ex: https://try.feldera.com)
        :param api_key: The optional API key to access Feldera
        :param version: The version of the API to use
        :param timeout: The timeout for the HTTP requests
        :param requests_verify: The `verify` parameter passed to the requests
            library. `True` by default.
        """

        self.url: str = url
        self.api_key: Optional[str] = api_key
        self.version: Optional[str] = version or "v0"
        self.timeout: Optional[float] = timeout
        self.requests_verify: bool = requests_verify
